/* sd_dynalloc.h 
 * dynamically allocated structures for the sd driver 
 * (c) Kurt Garloff <garloff@suse.de>, 7/2002
 * License: GNU GPL v2
 */

#ifndef _SD_DYNALLOC_H
#define _SD_DYNALLOC_H

#include <linux/blk.h>
#include "sd.h"

/* Should default to the officially assigned 16 */
#define MAX_SCSI_DYN_MAJORS CONFIG_SD_MAX_MAJORS
#define MINORS (1U<<MINORBITS)
#define MAJORS 256


#define MIDX_TO_MAJOR(i)	(sd_midx_to_major(i))
#define MAJOR_TO_MIDX(m)	(sd_majors? sd_major_to_midx[m]: SD_NO_MIDX)

#define DIDX_TO_MINOR(di)	((di)<<SCSI_DISK_PART_BITS)
#define MINOR_TO_DIDX(mn)	((mn)>>SCSI_DISK_PART_BITS)

#define DISKNO_TO_MIDX(i)	((i)>>SCSI_DISK_DISK_SHIFT)
#define DISKNO_TO_DIDX(i)	((i) & (SCSI_DISKS_PER_MAJOR-1))
#define DISKNO_TO_MAJOR(i)	(MIDX_TO_MAJOR(DISKNO_TO_MIDX(i)))
#define DISKNO_TO_MINOR(i)	(DIDX_TO_MINOR(DISKNO_TO_DIDX(i)))

#define MDIDX_TO_DISKNO(mi,di)	(((mi)<<SCSI_DISK_DISK_SHIFT) + (di))
#define MAJMIN_TO_DISKNO(mj,mn)	MDIDX_TO_DISKNO(MAJOR_TO_MIDX(mj),\
					(mn)>>SCSI_DISK_PART_BITS)
#define KDEV_TO_DISKNO(dev)	MAJMIN_TO_DISKNO(MAJOR(dev),MINOR(dev))

#define MDIDX_TO_KDEV(mi,di)	(MKDEV(MIDX_TO_MAJOR(mi),DIDX_TO_MINOR(di)))


#define SD_MAJOR_PTR(i)		(sd_major_ptrs[i])
#define SD_DISK_PTR(sdm,i)	(&((sdm)->disks[i]))

/* It would have made more sense to have per_partition structures,
 * embedded into per_disk structures embedded into per_major
 * structures. However, this would not have fit into the gendisk
 * scheme. Therefore have these arrays in per_major struct */

typedef struct sd_per_major {
	int  sd_sizes[MINORS];		/* size of partitions */
	int  sd_blocksizes[MINORS];	/* fs block size */
	int  sd_hardsizes[MINORS];	/* hw block size */
	int  sd_max_sectors[MINORS];	/* max request size in sectors */
	char sd_varyio[MINORS];
	
	Scsi_Disk disks[SCSI_DISKS_PER_MAJOR];
	devfs_handle_t	de_arr[SCSI_DISKS_PER_MAJOR];
	char flags[SCSI_DISKS_PER_MAJOR];
	struct gendisk	 sd_gendisk;
	struct hd_struct sd_hds[MINORS];
} SD_Major;

/* The dynamically allocated majors */
extern int sd_majors;
extern unsigned char sd_major_to_midx[MAJORS];
extern struct sd_per_major *sd_major_ptrs[MAX_SCSI_DYN_MAJORS];

#define SD_NO_MIDX 255
#define SD_NO_MAJOR -1

extern rwlock_t sd_dyn_lock;
#define SD_LOCK_W write_lock(&sd_dyn_lock)
#define SD_UNLOCK_W write_unlock(&sd_dyn_lock)

int  sd_midx_to_major  (const int midex);
int  sd_alloc_major    (void);
void sd_dealloc_major  (const int idx);
void sd_deallocate_all (void);
int  sd_find_free_slot (Scsi_Device *SDp, int *midx, int *didx);


#endif	/* _SD_DYNALLOC_H */
