/*
 * IBM Dynamic Probes
 * Copyright (c) International Business Machines Corp., 2000
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <linux/hook.h>
#include <linux/dprobes.h>

void dprobes_hooks_asm_code_start(void) { }

USE_HOOK(DO_TRAP);
USE_HOOK(DO_IABR);
USE_HOOK(DO_PGCK);
USE_HOOK(DO_DEBUG);
USE_HOOK(DO_PAGE_FAULT);
USE_HOOK(DPROBES_CALLK);

static int do_trap_exit(struct hook *h, struct pt_regs * regs)
{
	dp_handle_fault(regs);
	return HOOK_CONTINUE;
}	


static int do_iabr_exit(struct hook *h, int *rc, struct pt_regs * regs)
{
	if ((*rc = dp_trap(regs, DP_PROBE_BREAKPOINT, 0, regs->nip)))
		return HOOK_RETURN;
	return HOOK_CONTINUE;
}

static int do_pgck_exit(struct hook *h, int *rc, struct pt_regs * regs)
{
	if ((*rc = dp_trap(regs, DP_PROBE_BREAKPOINT, 0, regs->nip)))
		return HOOK_RETURN;
	return HOOK_CONTINUE;
}

static int do_debug_exit(struct hook *h, struct pt_regs * regs)
{
	if (dp_trap1(regs))
		return HOOK_RETURN;
	return HOOK_CONTINUE;
}

static int do_page_fault_exit(struct hook *h, struct pt_regs * regs)
{
	if (dp_pf(regs))
		return HOOK_RETURN;
	return HOOK_CONTINUE;
}

struct hook_rec do_trap_hook_rec;
struct hook_rec do_iabr_hook_rec;
struct hook_rec do_pgck_hook_rec;
struct hook_rec do_debug_hook_rec;
struct hook_rec do_page_fault_hook_rec;

int register_asm_dprobes_hooks(void)
{
	int retval;
	
	do_trap_hook_rec.hook_exit = do_trap_exit;
	if ((retval = hook_exit_register(&DO_TRAP, &do_trap_hook_rec)))
		goto err;

	do_pgck_hook_rec.hook_exit = do_pgck_exit;
	if ((retval = hook_exit_register(&DO_PGCK, &do_pgck_hook_rec)))
		goto err1;

	do_iabr_hook_rec.hook_exit = do_iabr_exit;
	if ((retval = hook_exit_register(&DO_IABR, &do_iabr_hook_rec)))
		goto err2;

	do_debug_hook_rec.hook_exit = do_debug_exit;
	if ((retval = hook_exit_register(&DO_DEBUG, &do_debug_hook_rec)))
		goto err3;

	do_page_fault_hook_rec.hook_exit = do_page_fault_exit;
	if ((retval = hook_exit_register(&DO_PAGE_FAULT, &do_page_fault_hook_rec)))
		goto err4;
	return 0;
	
err4:	hook_exit_deregister(&do_debug_hook_rec);
err3:	hook_exit_deregister(&do_iabr_hook_rec);
err2:	hook_exit_deregister(&do_pgck_hook_rec);
err1:	hook_exit_deregister(&do_trap_hook_rec);
err:	return retval;

}

int deregister_asm_dprobes_hooks(void)
{
	hook_exit_deregister(&do_page_fault_hook_rec);
	hook_exit_deregister(&do_debug_hook_rec);
	hook_exit_deregister(&do_iabr_hook_rec);
	hook_exit_deregister(&do_pgck_hook_rec);
	hook_exit_deregister(&do_trap_hook_rec);
	return 0;
}

int arm_all_asm_dprobes_hooks(void)
{
	hook_exit_arm(&do_trap_hook_rec);
	hook_exit_arm(&do_pgck_hook_rec);
	hook_exit_arm(&do_iabr_hook_rec);
	hook_exit_arm(&do_debug_hook_rec);
	hook_exit_arm(&do_page_fault_hook_rec);
	return 0;
}

int disarm_all_asm_dprobes_hooks(void)
{
	hook_exit_disarm(&do_page_fault_hook_rec);
	hook_exit_disarm(&do_debug_hook_rec);
	hook_exit_disarm(&do_iabr_hook_rec);
	hook_exit_disarm(&do_pgck_hook_rec);
	hook_exit_disarm(&do_trap_hook_rec);
	return 0;
}

void dprobes_hooks_asm_code_end(void) { }
