/* sg_dynalloc.h 
 * dynamically allocated structures for the sg driver 
 * derived by Holger Smolinski <holger@kunterbunt.bb.bawue.de>
 * from
 * sd_dynalloc: (c) Kurt Garloff <garloff@suse.de>, 7/2002
 * License: GNU GPL v2
 */

#ifndef _SG_DYNALLOC_H
#define _SG_DYNALLOC_H

#include "sg_base.h"

#define MAX_SG_DYN_MAJORS 32
#define MAJORS 256
#define SG_DEVICES_PER_MAJOR (1U << MINORBITS)

#define SG_MIDX_TO_MAJOR(i)	(sg_midx_to_major(i))
#define SG_MAJOR_TO_MIDX(m)	(sg_majors? sg_major_to_midx[m]: SG_NO_MIDX)

#define SG_MAJOR_PTR(i)		(sg_major_ptrs[i])
#define SG_DEVICE_PTR(sgm,i)	(&(sgm)->devices[i])

/* It would have made more sense to have per_partition structures,
 * embedded into per_disk structures embedded into per_major
 * structures. However, this would not have fit into the gendisk
 * scheme. Therefore have these arrays in per_major struct */

typedef struct sg_per_major {
	int major;
	Sg_device *devices[SG_DEVICES_PER_MAJOR];
} SG_Major;

/* The dynamically allocated majors */
extern int sg_majors;
extern unsigned char sg_major_to_midx[MAJORS];
extern SG_Major *sg_major_ptrs[MAX_SG_DYN_MAJORS];
#define SG_NO_MIDX 255
#define SG_NO_MAJOR -1

extern rwlock_t sg_dyn_lock;
#define SG_LOCK_W write_lock(&sg_dyn_lock)
#define SG_UNLOCK_W write_unlock(&sg_dyn_lock)

int  sg_midx_to_major  (const int midex);
int  sg_alloc_major    (void);
void sg_dealloc_major  (const int idx);
void sg_deallocate_all (void);
int  sg_find_free_slot (Scsi_Device *SGp, int *midx, int *didx);


#endif	/* _SG_DYNALLOC_H */
