/*
 *  fs/partitions/xbox.c
 *
 *  //TODO: find and insert GPL notice
 *
 *  June 2002 by SpeedBump:	initial implementation
 */

#include <linux/config.h>
#include <linux/delay.h>
#include <linux/fs.h>
#include <linux/genhd.h>
#include <linux/kernel.h>
#include <linux/major.h>
#include <linux/string.h>
#include <linux/blk.h>

#ifdef CONFIG_BLK_DEV_IDE
#include <linux/ide.h>	/* IDE xlate */
#endif /* CONFIG_BLK_DEV_IDE */

#include <asm/system.h>

#include "check.h"
#include "xbox.h"

#define XBOX_SECTOR_STORE	(0x0055F400L)
#define XBOX_SECTOR_SYSTEM	(0x00465400L)
#define XBOX_SECTOR_CONFIG	(0x00000000L)
#define XBOX_SECTOR_CACHE1	(0x00000400L)
#define XBOX_SECTOR_CACHE2	(0x00177400L)
#define XBOX_SECTOR_CACHE3	(0x002EE400L)
//#define XBOX_SECTOR_EXTEND	(0x00EEEC00L)
#define XBOX_SECTOR_EXTEND	(0x00EE8AB0L)
#define XBOX_SECTORS_CONFIG	(XBOX_SECTOR_CACHE1 - XBOX_SECTOR_CONFIG)

#define XBOX_SECTORS_STORE	(XBOX_SECTOR_EXTEND - XBOX_SECTOR_STORE)
#define XBOX_SECTORS_SYSTEM	(XBOX_SECTOR_STORE  - XBOX_SECTOR_SYSTEM)
#define XBOX_SECTORS_CACHE1	(XBOX_SECTOR_CACHE2 - XBOX_SECTOR_CACHE1)
#define XBOX_SECTORS_CACHE2	(XBOX_SECTOR_CACHE3 - XBOX_SECTOR_CACHE2)
#define XBOX_SECTORS_CACHE3	(XBOX_SECTOR_SYSTEM - XBOX_SECTOR_CACHE3)
#define XBOX_SECTORS_CONFIG	(XBOX_SECTOR_CACHE1 - XBOX_SECTOR_CONFIG)

#define XBOX_SECTOR_MAGIC	(3L)

static int
xbox_sig_string_match(	struct block_device *bdev, 
			unsigned long at_sector,
			char *expect )
{
	Sector sect;
	int retv;
	char *data;

	data = read_dev_sector(bdev, at_sector, &sect);
	
	if (!data) return 0;
	
	if (*(u32*)expect == *(u32*)data) retv = 1; else retv = 0;
	
	put_dev_sector(sect);
	
	/*
	if (!retv) {
		printk("xbox_sig_string_match: %s not found...found %c%c%c%c\n",
			expect,data[0],data[1],data[2],data[3]);
		for(i = 1; i<=512; i++) {
			printk(((i%32)?"%02X ":"%02X\n"),(unsigned char)data[i]);
		}
	}
	*/
	return retv;
}

static inline int
xbox_drive_detect(struct block_device *bdev)
{
	
	/** 
	* "BRFR" is apparently the magic number in the config area
	* the others are just paranoid checks to assure the expected
	* "FATX" tags for the other xbox partitions
	*
	* the odds against a non-xbox drive having random data to match is
	* astronomical...but it's possible I guess...you should only include
	* this check if you actually *have* an xbox drive...since it has to
	* be detected first
	*
	* @see check.c
	*/
	if (	(xbox_sig_string_match(bdev,XBOX_SECTOR_MAGIC ,"BRFR")) &&
		(xbox_sig_string_match(bdev,XBOX_SECTOR_SYSTEM,"FATX")) &&
		(xbox_sig_string_match(bdev,XBOX_SECTOR_STORE ,"FATX"))) {
		return 1; //success
	}
	
	return 0; // no xbox drive
}

int xbox_partition(struct gendisk *hd, struct block_device *bdev,
		   unsigned long first_sector, int first_part_minor)
{
	ide_drive_t *drive;
	unsigned long last_sector;
	unsigned long last_lba28_sector;
	int minor = first_part_minor;
	int retv;
	
	// return if not hda, avoiding NULL pointers and Oopses
	if (hd->major != 3) return 0;

	drive = get_info_ptr(to_kdev_t(bdev->bd_dev));
	last_sector = drive->capacity - 1;

	if (first_sector != 0) {
		//we only accept whole ide drives...no partials
		printk("xbox_partition: failed...first_sector != 0 == %ld\n",first_sector);
		return 0;
	}
	
	retv = xbox_drive_detect(bdev);
	if (retv > 0) {
		/* trying to find the first free partition */
		minor = 50;
		add_gd_partition(hd,minor++,XBOX_SECTOR_STORE ,XBOX_SECTORS_STORE );
		add_gd_partition(hd,minor++,XBOX_SECTOR_SYSTEM,XBOX_SECTORS_SYSTEM);
		add_gd_partition(hd,minor++,XBOX_SECTOR_CACHE1,XBOX_SECTORS_CACHE1);
		add_gd_partition(hd,minor++,XBOX_SECTOR_CACHE2,XBOX_SECTORS_CACHE2);
		add_gd_partition(hd,minor++,XBOX_SECTOR_CACHE3,XBOX_SECTORS_CACHE3);
		//there are some questions about the usage of any "extra" sectors...
		//I'll map everything after the end to an aditional sector.
		//
		//  Or maybe not.  You see, the F partition used by non-linux 
		//  operating systems usually can't extend past the 137 Gb limit
		//  imposed by 28-bit LBA.  So, let's make F only extend up to that
		//  point.  Extra space on disk will go to a linux-only hda56.
		if (XBOX_SECTOR_EXTEND < last_sector) {
			if ( last_sector > 0x0FFFFFFFL )  {
				last_lba28_sector = 0x0FFFFFFFL;
			} else {
				last_lba28_sector = last_sector;
			}
			if (xbox_sig_string_match(bdev,XBOX_SECTOR_EXTEND ,"FATX")) {
				add_gd_partition(hd, 
						minor++,
						XBOX_SECTOR_EXTEND,
						last_lba28_sector - XBOX_SECTOR_EXTEND);
			} else {
				minor++;
			}

			//  Now we create hda56.  Whether 55 exists or not, this will 
			//  always be hda56.  It only exists if there are sectors in-
			//  accessible to retail-style bioses.

			if ( last_sector > last_lba28_sector ) {
				add_gd_partition(hd, 
						minor++,
						last_lba28_sector,
						last_sector - last_lba28_sector);
			}

		}
	} else {
		//not an xbox drive
		return 0;
	}
}
