/*
 * Kernel header file for Linux crash dumps.
 *
 * Created by: Matt Robinson (yakker@sgi.com)
 *
 * Copyright 1999 Silicon Graphics, Inc. All rights reserved.
 * 
 * This code is released under version 2 of the GNU GPL.
 */

/* This header file holds the architecture specific crash dump header */
#ifndef _ASM_DUMP_H
#define _ASM_DUMP_H

/* necessary header files */
#include <asm/ptrace.h>                          /* for pt_regs             */

/* definitions */
#define DUMP_ASM_MAGIC_NUMBER     0xdeaddeadULL  /* magic number            */
#define DUMP_ASM_VERSION_NUMBER   0x2            /* version number          */


/*
 * Structure: dump_header_asm_t
 *  Function: This is the header for architecture-specific stuff.  It
 *            follows right after the dump header.
 */
typedef struct _dump_header_asm_s {

        /* the dump magic number -- unique to verify dump is valid */
        uint64_t             dha_magic_number;

        /* the version number of this dump */
        uint32_t             dha_version;

        /* the size of this header (in case we can't read it) */
        uint32_t             dha_header_size;

	/* the stack pointer for Alpha systems */
	uint64_t             dha_sp;

	/* the return address for Alpha systems */
	uint64_t             dha_ra;

	/* the dump registers */
	struct pt_regs       dha_regs;

} dump_header_asm_t;

#ifdef __KERNEL__
#define get_current()                                           \
({                                                              \
        register unsigned long sp;                              \
        asm("bis $30,$30,%0" : "=r" (sp));                      \
        sp;                                                     \
})          
#endif

#endif /* _ASM_DUMP_H */
