/*
 * Initialize IPSEC4
 *
 * Copyright (C) 2003 SuSE Linux AG
 *
 * Written by okir@suse.de
 */

#include <linux/config.h>
#include <linux/version.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>

#include <linux/ipsec.h>
#include <net/sadb.h>
#include <net/spd.h>

#include "ipsec_tunnel.h"
#include "ipsec4.h"

static struct inet_protocol ah_protocol = {
	handler: ipsec4_rcv,
	protocol: IPPROTO_AH,
	name: "AH"
};

static struct inet_protocol esp_protocol = {
	handler: ipsec4_rcv,
	protocol: IPPROTO_ESP,
	name: "ESP"
};

static int __init
ipsec4_init(void)
{
	int err;

	IPSEC4_DEBUG("Initializing ipsec4\n");
	printk("Initalizing IPsec4; based on FreeSWAN 1.98b\n");

	if ((err < ipsec4_tunnel_init()) < 0)
		return err;

	ipsec4_proc_init();

	inet_add_protocol(&esp_protocol);
	inet_add_protocol(&ah_protocol);

	return 0;
}

static void __exit
ipsec4_cleanup(void)
{
	inet_del_protocol(&esp_protocol);
	inet_del_protocol(&ah_protocol);
	ipsec4_proc_cleanup();
	ipsec4_tunnel_cleanup();
}

module_init(ipsec4_init);
module_exit(ipsec4_cleanup);
MODULE_LICENSE("Dual BSD/GPL");
