/*
 * General IPsec4 related functions
 *
 * Copyright (C) 2003 SuSE Linux AG
 *
 * Written by okir@suse.de
 */


#ifndef _IPSEC4_H_
#define _IPSEC4_H_

#define AH_AMAX	32
#define ESP_IVMAX 32

extern int		ipsec4_compute_ah(struct auth_algo_info *,
				struct sk_buff *,
				unsigned int, unsigned char *, size_t);
extern struct ipsec_sa *ipsec4_policy_bind(struct ipsec_sp *,
				struct sa_index **);
extern int		ipsec4_policy_to_salist(struct ipsec_sp *,
				struct ipsec_sa **,
				unsigned int);
extern struct ipsec_sp *ipsec4_policy_lookup(struct iphdr *, int);
extern int		ipsec4_acquire(struct ipsec_sp *, struct sk_buff *);
extern int		ipsec4_new_shunt(struct iphdr *, int,
				struct ipsec_sp **);

extern int		ipsec4_tunnel_init(void);
extern void		ipsec4_tunnel_cleanup(void);
extern int		ipsec4_proc_init(void);
extern void		ipsec4_proc_cleanup(void);

extern int		ipsec4_rcv(struct sk_buff *);
extern struct net_device *ipsec4_tunnel_from_device(struct net_device *);


#ifdef CONFIG_IPSEC_DEBUG
extern void		ipsec4_print_ip(const char *, struct iphdr *);
extern char *		ipsec4_print_sa(struct ipsec_sa *, char *, size_t);
#endif

#ifdef CONFIG_IPSEC_DEBUG
extern int sysctl_ipsec_debug_ipv4;

#define IPSEC4_ERROR(fmt, args...) \
	printk(KERN_NOTICE "%s: " fmt, __FUNCTION__, ##args)
#define IPSEC4_DEBUG(fmt, args...) \
	do { if (sysctl_ipsec_debug_ipv4) \
	printk(KERN_NOTICE "%s: " fmt, __FUNCTION__, ##args); \
	} while (0)
#define IPSEC4_PRINT_IP(iph) \
	do { if (sysctl_ipsec_debug_ipv4) \
	ipsec4_print_ip(__FUNCTION__, iph); \
	} while (0)
#else
#define IPSEC4_ERROR(fmt...) do { } while (0)
#define IPSEC4_DEBUG(fmt...) do { } while (0)
#define IPSEC4_PRINT_IP(iph) do { } while (0)
#endif

#endif /* _IPSEC4_H_ */
