/*
 * We need /proc/net/ipsec_version because a bunch of freeswan
 * scripts insist on checking this file. Also provide ipsec_tncfg
 * for the WLAN ipsec project.
 *
 * Written by Olaf Kirch <okir@suse.de>, Kurt Garloff <garloff@suse.de>
 */

#include <linux/config.h>
#include <linux/version.h>
#include <linux/module.h>
#include <linux/kernel.h>
#ifdef CONFIG_PROC_FS
#include <linux/proc_fs.h>
#endif

#include <linux/ipsec.h>
#include <net/sadb.h>
#include <net/spd.h>
#include "ipsec4.h"
#include "ipsec_tunnel.h"

static int
ipsec_version_get_info(char *buf, char **start, off_t offset, int count)
{
	int n;

	n = sprintf(buf, "USAGI IPsec4; based on FreeSWAN 1.98b\n");
	*start = buf + offset;
	n -= offset;
	if (n > count)
		n = count;
	return n;
}

static int prn_ipsec_dev_tncfg(struct net_device *dev, char* buffer)
{
	struct net_device *privdev;
	struct ipsecpriv *priv;
	int len = 0;
	priv = (struct ipsecpriv *)(dev->priv);
	len += sprintf(buffer + len, "%s",
		       dev->name);
	if (priv) {
		privdev = (struct net_device *)(priv->dev);
		len += sprintf(buffer + len, " -> %s",
			       privdev ? privdev->name : "NULL");
		len += sprintf(buffer + len, " mtu=%d(%d) -> %d",
			       dev->mtu,
			       priv->mtu,
			       privdev ? privdev->mtu : 0);
	}
	len += sprintf(buffer + len, "\n");
	return len;
}

static int
ipsec_tncfg_get_info(char* buf, char **start, off_t offset, int count)
{
	int len = 0;
	
	len += prn_ipsec_dev_tncfg(&dev_ipsec0, buf+len);
	len += prn_ipsec_dev_tncfg(&dev_ipsec1, buf+len);
	len += prn_ipsec_dev_tncfg(&dev_ipsec2, buf+len);
	len += prn_ipsec_dev_tncfg(&dev_ipsec3, buf+len);
	
	*start = buf + offset;		/* Start of wanted data */
	len -= offset;			/* Start slop */
	if (len > count)
		len = count;
	return len;
}

int
ipsec4_proc_init(void)
{
	proc_net_create("ipsec_version", 0, ipsec_version_get_info);
	proc_net_create("ipsec_tncfg", 0, ipsec_tncfg_get_info);
	return 0;
}

void
ipsec4_proc_cleanup(void)
{
	proc_net_remove("ipsec_tncfg");
	proc_net_remove("ipsec_version");
}
