/*
 *      Mobile-node header file
 *
 *      Authors:
 *      Sami Kivisaari          <skivisaa@cc.hut.fi>
 *
 *      $Id: s.mn.h 1.48 02/12/19 16:14:47+02:00 antti@jon.mipl.mediapoli.com $
 *
 *      This program is free software; you can redistribute it and/or
 *      modify it under the terms of the GNU General Public License
 *      as published by the Free Software Foundation; either version
 *      2 of the License, or (at your option) any later version.
 *
 */

#ifndef _MN_H
#define _MN_H

#include <linux/config.h>
#include <linux/in6.h>
#include <net/mipv6.h>

#include "sortedlist.h"
#include "mipv6_icmp.h"
#include "prefix.h"
#ifdef IPV6_MOBILE_NODE

/* constants for sending of BUs*/
#define HA_BU_DELAY 0 

#define HA_BU_DEF_LIFETIME 10000
#define CN_BU_DELAY 500 /* In case there is some data pckets to add the bu to */
#define CN_BU_DEF_LIFETIME 420 /* Max lifetime for RR bindings from mipv6 draft v19 */  
#define DUMB_CN_BU_LIFETIME 600 /* BUL entry lifetime in case of dumb CN */
#define ROUTER_BU_DEF_LIFETIME 30 /* For packet forwarding from previous coa */

#define ERROR_DEF_LIFETIME DUMB_CN_BU_LIFETIME

//#define CN_REQ_ACK /* Request Acks from CN's */

/*
 * Global configuration flags
 */
extern int mipv6_is_mn;

extern rwlock_t mn_info_lock;

#define MN_NOT_AT_HOME 0
#define MN_RETURNING_HOME 1
#define MN_AT_HOME 2

/*
 * Mobile Node information record
 */
struct mn_info {
	struct in6_addr home_addr;
	struct in6_addr ha;
	__u8 home_plen;
	__u8 is_at_home;
	__u8 has_home_reg;
	__u8 man_conf;
	int ifindex;
	unsigned long home_addr_expires;
	unsigned short dhaad_id;
	struct list_head list;
	spinlock_t lock;
};

/* prototypes for interface functions */
int mipv6_mn_init(void);
void mipv6_mn_exit(void);


struct handoff;
/* Interface to source address selection in IPv6 */
void mipv6_get_saddr_hook(struct inet6_ifaddr *ifp,
			  struct in6_addr *homeaddr);

/* Interface to movement detection */
int mipv6_mobile_node_moved(struct handoff *ho);

void mipv6_mn_send_home_na(struct in6_addr *haddr);
/* Init home reg. with coa */
int init_home_registration(struct in6_addr *home_addr, struct in6_addr *coa);

/* mn_info functions that require locking by caller */
struct mn_info *mipv6_mninfo_get_by_home(struct in6_addr *haddr);

struct mn_info *mipv6_mninfo_get_by_ha(struct in6_addr *home_agent);

struct mn_info *mipv6_mninfo_get_by_id(unsigned short id);

/* "safe" mn_info functions */
void mipv6_mninfo_add(int ifindex, struct in6_addr *home_addr, int plen, 
		      int isathome, unsigned long lifetime, struct in6_addr *ha, 
		      int ha_plen, unsigned long ha_lifetime, int man_conf);

int mipv6_mninfo_del(struct in6_addr *home_addr, int del_dyn_only);

int mipv6_mn_is_at_home(struct in6_addr *addr);

int mipv6_mn_is_home_addr(struct in6_addr *addr);

__u32 mipv6_mn_get_bulifetime(struct in6_addr *home_addr, 
			      struct in6_addr *coa, __u8 flags);

int mipv6_mn_may_solicit_ha(struct in6_addr *target);

int mipv6_mn_may_advertise(struct in6_addr *target, struct in6_addr *saddr);

void
mipv6_mn_prefix_route(struct in6_addr *pfx, int plen, int delete, 
		      unsigned long expires);

int mipv6_mn_use_hao(struct in6_addr *daddr, struct in6_addr *saddr);

void mipv6_check_tunneled_packet(struct sk_buff *skb);
/*
 * This function is deprecated and they should not be used, but some
 * parts still use them.  It only works for one home address.
 */
int mipv6_mn_get_homeaddr(struct in6_addr *home_addr);

/**
 * set_mn_pfx_list - manipulate pfx_list for MN when timer goes off
 * @entry: pfx_list_entry that is due
 */
static inline void set_mn_pfx_list(struct pfx_list_entry *entry)
{
	extern struct list_head pfx_list;

	mipv6_icmpv6_send(&entry->daddr, &entry->saddr, MIPV6_PREFIX_SOLICIT, 
			  0, NULL, NULL, 0); /* entry->ifindex */

	if (++entry->retries <= MAX_RTR_SOLICITATIONS)
		mipv6_slist_push_first(&pfx_list, jiffies +
				user_to_kernel_hz_overflow(RTR_SOLICITATION_INTERVAL));
	else
		kfree(mipv6_slist_del_first(&pfx_list));
}

#else	/* IPV6_MOBILE_NODE */

static __inline__ void set_mn_pfx_list(struct pfx_list_entry *entry)
{ return; }
static inline int mipv6_mn_use_hao(struct in6_addr *daddr, 
				   struct in6_addr *saddr)
{ return 0; }
static inline void mipv6_check_tunneled_packet(struct sk_buff *skb)
{ return; }

#endif	/* IPV6_MOBILE_NODE */

#endif /* _MN_H */
